# energyflow needs to be imported before pytorch
# because of https://github.com/pkomiske/EnergyFlow/issues/24
from energyflow.emd import emds

import logging
import warnings
from typing import Union, Tuple

import numpy as np
import torch
from torch import Tensor
from torch.utils.data import DataLoader

from jetnet.datasets import JetNet
from jetnet import utils

from scipy import linalg
from scipy.stats import wasserstein_distance

# from .particlenet import ParticleNet

from tqdm import tqdm

import pathlib
import sys


rng = np.random.default_rng()


# TODOs:
# - Functionality for adding FPND for new datasets
# - Cartesian coordinates


def _optional_tqdm(iter_obj, use_tqdm, total=None, desc=None):
    if use_tqdm:
        return tqdm(iter_obj, total=total, desc=desc)
    else:
        return iter_obj


# from https://github.com/mseitzer/pytorch-fid
def _calculate_frechet_distance(mu1, sigma1, mu2, sigma2, eps=1e-6):
    """Numpy implementation of the Frechet Distance.
    The Frechet distance between two multivariate Gaussians X_1 ~ N(mu_1, C_1)
    and X_2 ~ N(mu_2, C_2) is
            d^2 = ||mu_1 - mu_2||^2 + Tr(C_1 + C_2 - 2*sqrt(C_1*C_2)).
    Stable version by Dougal J. Sutherland.
    Params:
    -- mu1   : Numpy array containing the activations of a layer of the
               inception net (like returned by the function 'get_predictions')
               for generated samples.
    -- mu2   : The sample mean over activations, precalculated on an
               representative data set.
    -- sigma1: The covariance matrix over activations for generated samples.
    -- sigma2: The covariance matrix over activations, precalculated on an
               representative data set.
    Returns:
    --   : The Frechet Distance.
    """

    mu1 = np.atleast_1d(mu1)
    mu2 = np.atleast_1d(mu2)

    sigma1 = np.atleast_2d(sigma1)
    sigma2 = np.atleast_2d(sigma2)

    assert mu1.shape == mu2.shape, "Training and test mean vectors have different lengths"
    assert sigma1.shape == sigma2.shape, "Training and test covariances have different dimensions"

    diff = mu1 - mu2

    # Product might be almost singular
    covmean, _ = linalg.sqrtm(sigma1.dot(sigma2), disp=False)
    if not np.isfinite(covmean).all():
        msg = (
            "fid calculation produces singular product; " "adding %s to diagonal of cov estimates"
        ) % eps
        logging.debug(msg)
        offset = np.eye(sigma1.shape[0]) * eps
        covmean = linalg.sqrtm((sigma1 + offset).dot(sigma2 + offset))

    # Numerical error might give slight imaginary component
    if np.iscomplexobj(covmean):
        if not np.allclose(np.diagonal(covmean).imag, 0, atol=1e-3):
            m = np.max(np.abs(covmean.imag))
            raise ValueError("Imaginary component {}".format(m))
        covmean = covmean.real

    tr_covmean = np.trace(covmean)

    return diff.dot(diff) + np.trace(sigma1) + np.trace(sigma2) - 2 * tr_covmean


_eval_module = sys.modules[__name__]
# for saving fpnd objects after the first loading
_eval_module.fpnd_dict = {"NUM_SAMPLES": 50000}


def _get_fpnd_real_mu_sigma(
    dataset_name: str,
    jet_type: str,
    num_particles: int,
    num_particle_features: int,
    data_dir: str,
    device: str = "cpu",
    batch_size: int = 16,
    use_tqdm: bool = True,
):
    """
    Get and save the statistics of ParticleNet activations on real jets.
    These should already come with the library so this method should not need to be run by the user.
    """
    from .particlenet import _ParticleNet

    _eval_module_path = str(pathlib.Path(__file__).parent.resolve())
    resources_path = f"{_eval_module_path}/fpnd_resources/{dataset_name}/{num_particles}_particles"

    pnet = _ParticleNet(num_particles, num_particle_features).to(device)
    pnet.load_state_dict(torch.load(f"{resources_path}/pnet_state_dict.pt", map_location=device))
    pnet.eval()

    if dataset_name == "jetnet":
        jets = JetNet(jet_type, data_dir, normalize=False, train=False, use_mask=False).data[
            : _eval_module.fpnd_dict["NUM_SAMPLES"]
        ]
        JetNet.normalize_features(jets, fpnd=True)
        # TODO other datasets
    else:
        raise RuntimeError("Only jetnet dataset implemented currently")

    # run inference and store activations
    jets_loaded = DataLoader(jets, batch_size)

    logging.info(f"Calculating ParticleNet activations on real jets with batch size {batch_size}")
    activations = []
    for i, jets_batch in _optional_tqdm(
        enumerate(jets_loaded), use_tqdm, total=len(jets_loaded), desc="Running ParticleNet"
    ):
        activations.append(pnet(jets_batch.to(device), ret_activations=True).cpu().detach().numpy())

    activations = np.concatenate(activations, axis=0)

    mu = np.mean(activations, axis=0)
    sigma = np.cov(activations, rowvar=False)

    np.savetxt(f"{resources_path}/{jet_type}_mu.txt", mu)
    np.savetxt(f"{resources_path}/{jet_type}_sigma.txt", sigma)


def _init_fpnd_dict(
    dataset_name: str,
    jet_type: str,
    num_particles: int,
    num_particle_features: int,
    device: str = "cpu",
):
    """Load the ParticleNet model and pre-saved statistics for real jets"""
    try:
        from .particlenet import _ParticleNet
    except ModuleNotFoundError:
        print("torch_geometric needs to be installed for FPND")
        raise

    if dataset_name not in _eval_module.fpnd_dict:
        _eval_module.fpnd_dict[dataset_name] = {}

    if num_particles not in _eval_module.fpnd_dict[dataset_name]:
        _eval_module.fpnd_dict[dataset_name][num_particles] = {}

    if jet_type not in _eval_module.fpnd_dict[dataset_name][num_particles]:
        _eval_module.fpnd_dict[dataset_name][num_particles][jet_type] = {}

    _eval_module_path = str(pathlib.Path(__file__).parent.resolve())
    resources_path = f"{_eval_module_path}/fpnd_resources/{dataset_name}/{num_particles}_particles"

    pnet = _ParticleNet(num_particles, num_particle_features)
    pnet.load_state_dict(torch.load(f"{resources_path}/pnet_state_dict.pt", map_location=device))

    _eval_module.fpnd_dict[dataset_name][num_particles][jet_type]["pnet"] = pnet
    _eval_module.fpnd_dict[dataset_name][num_particles][jet_type]["mu"] = np.loadtxt(
        f"{resources_path}/{jet_type}_mu.txt"
    )
    _eval_module.fpnd_dict[dataset_name][num_particles][jet_type]["sigma"] = np.loadtxt(
        f"{resources_path}/{jet_type}_sigma.txt"
    )


def fpnd(
    jets: Union[Tensor, np.ndarray],
    jet_type: str,
    dataset_name: str = "jetnet",
    device: str = None,
    batch_size: int = 16,
    use_tqdm: bool = True,
) -> float:
    """
    Calculates the Frechet ParticleNet Distance, as defined in https://arxiv.org/abs/2106.11535,
    for input ``jets`` of type ``jet_type``.

    ``jets`` are passed through our pretrained ParticleNet module and activations are compared
    with the cached activations from real jets.
    The recommended and max number of jets is 50,000.

    **torch_geometric must be installed separately for running inference with ParticleNet.**

    Currently FPND only supported for the JetNet dataset with 30 particles, but functionality for
    other datasets + ability for users to use their own version is in development.

    Args:
        jets (Union[Tensor, np.ndarray]): Tensor or array of jets, of shape
          ``[num_jets, num_particles, num_features]`` with features in order
          ``[eta, phi, pt, (optional) mask]``
        jet_type (str): jet type, out of ``['g', 't', 'q']``.
        dataset_name (str): Dataset to use. Currently only JetNet is supported.
          Defaults to "jetnet".
        device (str): 'cpu' or 'cuda'. If not specified, defaults to cuda if available else cpu.
        batch_size (int): Batch size for ParticleNet inference. Defaults to 16.
        use_tqdm (bool): use tqdm bar while getting ParticleNet activations. Defaults to True.

    Returns:
        float: the measured FPND.

    """
    assert dataset_name == "jetnet", "Only JetNet is currently supported with FPND"

    num_particles = jets.shape[1]
    num_particle_features = jets.shape[2]

    assert (
        num_particles == 30
    ), "Currently FPND only supported for 30 particles - more functionality coming soon."
    assert (
        num_particle_features == 3
    ), "Not the right number of particle features for the JetNet dataset."

    if jets.shape[0] < _eval_module.fpnd_dict["NUM_SAMPLES"]:
        warnings.warn(
            f"Recommended number of jets for FPND calculation is {_eval_module.fpnd_dict['NUM_SAMPLES']}",
            RuntimeWarning,
        )

    if isinstance(jets, np.ndarray):
        jets = Tensor(jets)

    jets = jets.clone()

    if device is None:
        device = "cuda" if torch.cuda.is_available() else "cpu"

    assert device == "cuda" or device == "cpu", "Invalid device type"

    if dataset_name == "jetnet":
        JetNet.normalize_features(jets, fpnd=True)
        # TODO other datasets

    # ParticleNet module and the real mu's and sigma's are only loaded once
    if (
        dataset_name not in _eval_module.fpnd_dict
        or num_particles not in _eval_module.fpnd_dict[dataset_name]
        or jet_type not in _eval_module.fpnd_dict[dataset_name][num_particles]
    ):
        _init_fpnd_dict(dataset_name, jet_type, num_particles, num_particle_features, device)

    pnet = _eval_module.fpnd_dict[dataset_name][num_particles][jet_type]["pnet"].to(device)
    pnet.eval()

    mu1 = _eval_module.fpnd_dict[dataset_name][num_particles][jet_type]["mu"]
    sigma1 = _eval_module.fpnd_dict[dataset_name][num_particles][jet_type]["sigma"]

    # run inference and store activations
    jets_loaded = DataLoader(jets[:, :, : _eval_module.fpnd_dict["NUM_SAMPLES"]], batch_size)

    logging.info(f"Calculating ParticleNet activations with batch size: {batch_size}")
    activations = []
    for i, jets_batch in _optional_tqdm(
        enumerate(jets_loaded), use_tqdm, total=len(jets_loaded), desc="Running ParticleNet"
    ):
        activations.append(pnet(jets_batch.to(device), ret_activations=True).cpu().detach().numpy())

    activations = np.concatenate(activations, axis=0)

    mu2 = np.mean(activations, axis=0)
    sigma2 = np.cov(activations, rowvar=False)

    fpnd = _calculate_frechet_distance(mu1, sigma1, mu2, sigma2)

    return fpnd


def w1p(
    jets1: Union[Tensor, np.ndarray],
    jets2: Union[Tensor, np.ndarray],
    mask1: Union[Tensor, np.ndarray] = None,
    mask2: Union[Tensor, np.ndarray] = None,
    exclude_zeros: bool = True,
    num_particle_features: int = 0,
    num_eval_samples: int = 10000,
    num_batches: int = 5,
    average_over_features: bool = True,
    return_std: bool = True,
):
    """
    Get 1-Wasserstein distances between particle features of ``jets1`` and ``jets2``.

    Args:
        jets1 (Union[Tensor, np.ndarray]): Tensor or array of jets, of shape
          ``[num_jets, num_particles_per_jet, num_features_per_particle]``.
        jets2 (Union[Tensor, np.ndarray]): Tensor or array of jets, of same format as ``jets1``.
        mask1 (Union[Tensor, np.ndarray]): Optional tensor or array of binary particle masks, of
          shape ``[num_jets, num_particles_per_jet]`` or ``[num_jets, num_particles_per_jet, 1]``.
          If given, 0-masked particles will be excluded from w1 calculation.
        mask2 (Union[Tensor, np.ndarray]): Optional tensor or array of same format as ``masks2``.
        exclude_zeros (bool): Ignore zero-padded particles i.e.
          those whose whose feature norms are exactly 0. Defaults to True.
        num_particle_features (int): Will return W1 scores of the first
          ``num_particle_features`` particle features. If 0, will calculate for all.
        num_eval_samples (int): Number of jets out of the total to use for W1 measurement.
          Defaults to 10000.
        num_batches (int): Number of different batches to average W1 scores over. Defaults to 5.
        average_over_features (bool): Average over the particle features to return
          a single W1-P score. Defaults to True.
        return_std (bool): Return the standard deviation as well of the W1 scores over the
          ``num_batches`` batches. Defaults to True.

    Returns:
        Tuple[Union[float, np.ndarray], Union[float, np.ndarray]]:
        - **Union[float, np.ndarray]**: if ``average_over_features`` is True, float of average W1
          scores for each particle feature, first averaged over ``num_batches``, else array of
          length ``num_particle_features`` containing average W1 scores for each feature.
        - **Union[float, np.ndarray]** `(optional, only if ``return_std`` is True)`: if
          ``average_over_features`` is True, float of standard deviation of all W1 scores for each
          particle feature, first calculated over ``num_batches`` then propagated for the final
          average, else array of length ``num_particle_features`` containing standard deviation W1
          scores for each feature.

    """
    assert len(jets1.shape) == 3 and len(jets2.shape) == 3, "input jets format is incorrect"

    if num_particle_features <= 0:
        num_particle_features = jets1.shape[2]

    assert (
        num_particle_features <= jets1.shape[2]
    ), "more particle features requested than were inputted"
    assert (
        num_particle_features <= jets2.shape[2]
    ), "more particle features requested than were inputted"

    if mask1 is not None:
        # TODO: should be wrapped in try catch
        mask1 = mask1.reshape(jets1.shape[0], jets1.shape[1])
        mask1 = mask1.astype(bool)

    if mask2 is not None:
        # TODO: should be wrapped in try catch
        mask2 = mask2.reshape(jets2.shape[0], jets2.shape[2])
        mask2 = mask2.astype(bool)

    if isinstance(jets1, Tensor):
        jets1 = jets1.cpu().detach().numpy()

    if isinstance(jets2, Tensor):
        jets2 = jets2.cpu().detach().numpy()

    if exclude_zeros:
        zeros1 = np.linalg.norm(jets1[:, :, :num_particle_features], axis=2) == 0
        mask1 = ~zeros1 if mask1 is None else mask1 * ~zeros1

        zeros2 = np.linalg.norm(jets2[:, :, :num_particle_features], axis=2) == 0
        mask2 = ~zeros2 if mask2 is None else mask2 * ~zeros2

    w1s = []

    for j in range(num_batches):
        rand1 = rng.choice(len(jets1), size=num_eval_samples)
        rand2 = rng.choice(len(jets2), size=num_eval_samples)

        rand_sample1 = jets1[rand1]
        rand_sample2 = jets2[rand2]

        if mask1 is not None:
            parts1 = rand_sample1[:, :, :num_particle_features][mask1[rand1]]
        else:
            parts1 = rand_sample1[:, :, :num_particle_features].reshape(-1, num_particle_features)

        if mask2 is not None:
            parts2 = rand_sample2[:, :, :num_particle_features][mask2[rand2]]
        else:
            parts2 = rand_sample2[:, :, :num_particle_features].reshape(-1, num_particle_features)

        if parts1.shape[0] == 0 or parts2.shape[0] == 0:
            w1 = [np.inf, np.inf, np.inf]
        else:
            w1 = [
                wasserstein_distance(parts1[:, i], parts2[:, i])
                for i in range(num_particle_features)
            ]

        w1s.append(w1)

    means = np.mean(w1s, axis=0)
    stds = np.std(w1s, axis=0)

    if average_over_features:
        return np.mean(means), np.linalg.norm(stds) if return_std else np.mean(means)
    else:
        return means, stds if return_std else means


def w1m(
    jets1: Union[Tensor, np.ndarray],
    jets2: Union[Tensor, np.ndarray],
    use_particle_masses: bool = False,
    num_eval_samples: int = 10000,
    num_batches: int = 5,
    return_std: bool = True,
):
    """
    Get 1-Wasserstein distance between masses of ``jets1`` and ``jets2``.

    Args:
        jets1 (Union[Tensor, np.ndarray]): Tensor or array of jets, of shape
          ``[num_jets, num_particles, num_features]`` with features in order
          ``[eta, phi, pt, (optional) mass]``
        jets2 (Union[Tensor, np.ndarray]): Tensor or array of jets, of same format as ``jets1``.
        num_eval_samples (int): Number of jets out of the total to use for W1 measurement.
          Defaults to 10000.
        num_batches (int): Number of different batches to average W1 scores over. Defaults to 5.
        return_std (bool): Return the standard deviation as well of the W1 scores over the
          ``num_batches`` batches. Defaults to True.

    Returns:
        Tuple[float, float]:
        - **float**: W1 mass score, averaged over ``num_batches``.
        - **float** `(optional, only if ``return_std`` is True)`: standard deviation of W1 mass
          scores over ``num_batches``.

    """
    assert len(jets1.shape) == 3 and len(jets2.shape) == 3, "input jets format is incorrect"

    if isinstance(jets1, Tensor):
        jets1 = jets1.cpu().detach().numpy()

    if isinstance(jets2, Tensor):
        jets2 = jets2.cpu().detach().numpy()

    masses1 = utils.jet_features(jets1)["mass"]
    masses2 = utils.jet_features(jets2)["mass"]

    w1s = []

    for j in range(num_batches):
        rand1 = rng.choice(len(masses1), size=num_eval_samples)
        rand2 = rng.choice(len(masses2), size=num_eval_samples)

        rand_sample1 = masses1[rand1]
        rand_sample2 = masses2[rand2]

        w1s.append(wasserstein_distance(rand_sample1, rand_sample2))

    return np.mean(w1s), np.std(w1s) if return_std else np.mean(w1s)


def w1efp(
    jets1: Union[Tensor, np.ndarray],
    jets2: Union[Tensor, np.ndarray],
    use_particle_masses: bool = False,
    efpset_args: list = [("n==", 4), ("d==", 4), ("p==", 1)],
    num_eval_samples: int = 10000,
    num_batches: int = 5,
    average_over_efps: bool = True,
    return_std: bool = True,
    efp_jobs: int = None,
):
    """
    Get 1-Wasserstein distances between Energy Flow Polynomials
    (Komiske et al. 2017 https://arxiv.org/abs/1712.07124) of ``jets1`` and ``jets2``.

    Args:
        jets1 (Union[Tensor, np.ndarray]): Tensor or array of jets of shape
          ``[num_jets, num_particles, num_features]``, with features in order
          ``[eta, phi, pt, (optional) mass]``. If no particle masses given
          (``particle_masses`` should be False), they are assumed to be 0.
        jets2 (Union[Tensor, np.ndarray]): Tensor or array of jets, of same format as ``jets1``.
        use_particle_masses (bool): Whether ``jets1`` and ``jets2`` have particle masses as their
          4th particle features. Defaults to False.
        efpset_args (List): Args for the energyflow.efpset function to specify which EFPs to use,
          as defined here https://energyflow.network/docs/efp/#efpset.
          Defaults to the n=4, d=5, prime EFPs.
        num_eval_samples (int): Number of jets out of the total to use for W1 measurement.
          Defaults to 10000.
        num_batches (int): Number of different batches to average W1 scores over. Defaults to 5.
        average_over_efps (bool): Average over the EFPs to return a single W1-EFP score.
          Defaults to True.
        return_std (bool): Return the standard deviation as well of the W1 scores over the
          ``num_batches`` batches. Defaults to True.
        efp_jobs (int): number of jobs to use for energyflow's EFP batch computation.
          None means as many processes as there are CPUs.

    Returns:
        Tuple[Union[float, np.ndarray], Union[float, np.ndarray]]:
        - **Union[float, np.ndarray]**: if ``average_over_features`` is True, float of average W1
          scores for each particle feature, first averaged over ``num_batches``, else array of
          length ``num_particle_features`` containing average W1 scores for each feature.
        - **Union[float, np.ndarray]** `(optional, only if ``return_std`` is True)`: if
          ``average_over_features`` is True, float of standard deviation of all W1 scores for each
          particle feature, first calculated over ``num_batches`` then propagated for the final
          average, else array of length ``num_particle_features`` containing standard deviation W1
          scores for each feature.

    """

    if isinstance(jets1, Tensor):
        jets1 = jets1.cpu().detach().numpy()

    if isinstance(jets2, Tensor):
        jets2 = jets2.cpu().detach().numpy()

    assert len(jets1.shape) == 3 and len(jets2.shape) == 3, "input jets format is incorrect"
    assert (jets1.shape[2] - int(use_particle_masses) >= 3) and (
        jets1.shape[2] - int(use_particle_masses) >= 3
    ), "particle feature format is incorrect"

    efps1 = utils.efps(
        jets1, use_particle_masses=use_particle_masses, efpset_args=efpset_args, efp_jobs=efp_jobs
    )
    efps2 = utils.efps(
        jets2, use_particle_masses=use_particle_masses, efpset_args=efpset_args, efp_jobs=efp_jobs
    )
    num_efps = efps1.shape[1]

    w1s = []

    for j in range(num_batches):
        rand1 = rng.choice(len(efps1), size=num_eval_samples)
        rand2 = rng.choice(len(efps2), size=num_eval_samples)

        rand_sample1 = efps1[rand1]
        rand_sample2 = efps2[rand2]

        w1 = [wasserstein_distance(rand_sample1[:, i], rand_sample2[:, i]) for i in range(num_efps)]
        w1s.append(w1)

    means = np.mean(w1s, axis=0)
    stds = np.std(w1s, axis=0)

    if average_over_efps:
        return np.mean(means), np.linalg.norm(stds) if return_std else np.mean(means)
    else:
        return means, stds if return_std else means


def cov_mmd(
    real_jets: Union[Tensor, np.ndarray],
    gen_jets: Union[Tensor, np.ndarray],
    num_eval_samples: int = 100,
    num_batches: int = 10,
    use_tqdm: bool = True,
) -> Tuple[float, float]:
    """
    Calculate coverage and MMD between real and generated jets,
    using the Energy Mover's Distance as the distance metric.

    Args:
        real_jets (Union[Tensor, np.ndarray]): Tensor or array of jets, of shape
          ``[num_jets, num_particles, num_features]`` with features in order ``[eta, phi, pt]``
        gen_jets (Union[Tensor, np.ndarray]): tensor or array of generated jets,
          same format as real_jets.
        num_eval_samples (int): number of jets out of the real and gen jets each between which to
          evaluate COV and MMD. Defaults to 100.
        num_batches (int): number of different batches to calculate COV and MMD and average over.
          Defaults to 100.
        use_tqdm (bool): use tqdm bar while calculating over ``num_batches`` batches.
          Defaults to True.

    Returns:
        Tuple[float, float]:
        - **float**: coverage, averaged over ``num_batches``.
        - **float**: MMD, averaged over ``num_batches``.

    """
    assert len(real_jets.shape) == 3 and len(gen_jets.shape) == 3, "input jets format is incorrect"
    assert (real_jets.shape[2] >= 3) and (
        gen_jets.shape[2] >= 3
    ), "particle feature format is incorrect"

    if isinstance(real_jets, Tensor):
        real_jets = real_jets.cpu().detach().numpy()

    if isinstance(gen_jets, Tensor):
        gen_jets = gen_jets.cpu().detach().numpy()

    assert np.all(real_jets[:, :, 2] >= 0) and np.all(
        gen_jets[:, :, 2] >= 0
    ), "particle pTs must all be >= 0 for EMD calculation"

    # convert from JetNet [eta, phi, pt] format to energyflow [pt, eta, phi]
    real_jets = real_jets[:, :, [2, 0, 1]]
    gen_jets = gen_jets[:, :, [2, 0, 1]]

    covs = []
    mmds = []

    for j in _optional_tqdm(
        range(num_batches), use_tqdm, desc=f"Calculating cov and mmd over {num_batches} batches"
    ):
        real_rand = rng.choice(len(real_jets), size=num_eval_samples)
        gen_rand = rng.choice(len(gen_jets), size=num_eval_samples)

        real_rand_sample = real_jets[real_rand]
        gen_rand_sample = gen_jets[gen_rand]

        # 2D array of emds, with shape (len(gen_rand_sample), len(real_rand_sample))
        dists = emds(gen_rand_sample, real_rand_sample)

        # for MMD, for each gen jet find the closest real jet and average EMDs
        mmds.append(np.mean(np.min(dists, axis=0)))

        # for coverage, for each real jet find the closest gen jet
        # and get the number of unique matchings
        covs.append(np.unique(np.argmin(dists, axis=1)).size / num_eval_samples)

    return np.mean(covs), np.mean(mmds)



from typing import Union, Tuple

import torch
from torch import nn, Tensor
import numpy as np


class EMDLoss(nn.Module):
    """
    Calculates the energy mover's distance between two batches of jets differentiably as a convex optimization problem
    either through the linear programming library ``cvxpy`` or by converting it to a quadratic programming problem and
    using the ``qpth`` library. ``cvxpy`` is marginally more accurate but ``qpth`` is significantly faster so defaults
    to ``qpth``.

    **JetNet must be installed with the extra option** ``pip install jetnet[emdloss]`` **to use this.**

    *Note: PyTorch <= 1.9 has a bug which will cause this to fail for >= 32 particles. This PR should fix this from 1.10 onwards*
    https://github.com/pytorch/pytorch/pull/61815.

    Args:
        method (str): 'cvxpy' or 'qpth'. Defaults to 'qpth'.
        num_particles (int): number of particles per jet - onlyneeds to be specified if method is 'cvxpy'.
        qpth_form (str): 'L2' or 'QP'. Defaults to 'L2'.
        qpth_l2_strength (float): regularization parameter for 'L2' qp form. Defaults to 0.0001.
        device (str): 'cpu' or 'cuda'. Defaults to 'cpu'.

    """

    def __init__(
        self,
        method: str = "qpth",
        num_particles: int = None,
        qpth_form: str = "L2",
        qpth_l2_strength: float = 0.0001,
        device: str = "cpu",
    ):
        super(EMDLoss, self).__init__()

        if method == "qpth":
            try:
                global qpth
                qpth = __import__("qpth", globals(), locals())
            except:
                print(
                    "QPTH needs to be installed separately to use this method - try pip install jetnet[emdloss]"
                )
                raise
        else:
            try:
                global cp, cvxpylayers
                cp = __import__("cvxpy", globals(), locals())
                cvxpylayers = __import__("cvxpylayers", globals(), locals())
            except:
                print(
                    "cvxpy needs to be installed separately to use this method - try pip install jetnet[emdloss]"
                )
                raise

        assert method == "qpth" or method == "cvxpy", "invalid method type"
        assert method != "cvxpy" or (
            num_particles is not None and num_particles > 0
        ), "num_particles must be specified to use 'cvxpy' method"
        assert qpth_form == "L2" or qpth_form == "QP", "invalid qpth form"
        assert device == "cpu" or device == "cuda", "invalid device type"

        self.num_particles = num_particles
        self.method = method
        if method == "qpth":
            self.form = qpth_form
            self.l2_strength = qpth_l2_strength
        self.device = device

        if method == "cvxpy":
            x = cp.Variable(num_particles * num_particles)  # flows
            c = cp.Parameter(num_particles * num_particles)  # costs
            w = cp.Parameter(num_particles + num_particles)  # weights
            Emin = cp.Parameter(1)  # min energy out of the two jets

            g1 = np.zeros((num_particles, num_particles * num_particles))
            for i in range(num_particles):
                g1[i, i * num_particles : (i + 1) * num_particles] = 1
            g2 = np.concatenate([np.eye(num_particles) for i in range(num_particles)], axis=1)
            g = np.concatenate((g1, g2), axis=0)

            constraints = [x >= 0, g @ x <= w, cp.sum(x) == Emin]
            objective = cp.Minimize(c.T @ x)
            problem = cp.Problem(objective, constraints)

            self.cvxpylayer = cvxpylayers.torch.CvxpyLayer(
                problem, parameters=[c, w, Emin], variables=[x]
            ).to(device)

    def _emd_inference_qpth(
        self, distance_matrix: Tensor, weight1: Tensor, weight2: Tensor
    ) -> Tuple[Tensor, Tensor]:
        """
        Using the QP solver QPTH to get EMDs (LP problem), adapted from https://github.com/icoz69/DeepEMD/blob/master/Models/models/emd_utils.py
        One can transform the LP problem to QP, or omit the QP term by multiplying it with a small value, i.e. l2_strngth.

        Args:
            distance_matrix (Tensor): nbatch * element_number * element_number.
            weight1 (Tensor): nbatch  * weight_number.
            weight2 (Tensor): nbatch  * weight_number.

        Returns:
            emd distance: nbatch*1
            flow : nbatch * weight_number *weight_number
        """
        nbatch = distance_matrix.shape[0]
        nelement_distmatrix = distance_matrix.shape[1] * distance_matrix.shape[2]
        nelement_weight1 = weight1.shape[1]
        nelement_weight2 = weight2.shape[1]

        # reshape dist matrix too (nbatch, 1, n1 * n2)
        Q_1 = distance_matrix.view(-1, 1, nelement_distmatrix).double()

        if (
            self.form == "QP"
        ):  # converting to QP - after testing L2 reg performs marginally better than QP
            # version: QTQ
            Q = torch.bmm(Q_1.transpose(2, 1), Q_1).double() + 1e-4 * torch.eye(
                nelement_distmatrix
            ).double().unsqueeze(0).repeat(
                nbatch, 1, 1
            )  # 0.00001 *
            p = torch.zeros(nbatch, nelement_distmatrix).double().to(self.device)
        elif self.form == "L2":  # regularizing a trivial Q term with l2_strength
            # version: regularizer
            Q = (
                (self.l2_strength * torch.eye(nelement_distmatrix).double())
                .unsqueeze(0)
                .repeat(nbatch, 1, 1)
                .to(self.device)
            )
            p = distance_matrix.view(nbatch, nelement_distmatrix).double()
        else:
            raise ValueError("Unkown form")

        # h = [0 ... 0 w1 w2]
        h_1 = torch.zeros(nbatch, nelement_distmatrix).double().to(self.device)
        h_2 = torch.cat([weight1, weight2], 1).double()
        h = torch.cat((h_1, h_2), 1)

        G_1 = (
            -torch.eye(nelement_distmatrix)
            .double()
            .unsqueeze(0)
            .repeat(nbatch, 1, 1)
            .to(self.device)
        )
        G_2 = (
            torch.zeros([nbatch, nelement_weight1 + nelement_weight2, nelement_distmatrix])
            .double()
            .to(self.device)
        )
        # sum_j(xij) = si
        for i in range(nelement_weight1):
            G_2[:, i, nelement_weight2 * i : nelement_weight2 * (i + 1)] = 1
        # sum_i(xij) = dj
        for j in range(nelement_weight2):
            G_2[:, nelement_weight1 + j, j::nelement_weight2] = 1

        # xij>=0, sum_j(xij) <= si,sum_i(xij) <= dj, sum_ij(x_ij) = min(sum(si), sum(dj))
        G = torch.cat((G_1, G_2), 1)
        A = torch.ones(nbatch, 1, nelement_distmatrix).double().to(self.device)
        b = torch.min(torch.sum(weight1, 1), torch.sum(weight2, 1)).unsqueeze(1).double()
        flow = qpth.qp.QPFunction(verbose=-1)(Q, p, G, h, A, b)

        energy_diff = torch.abs(torch.sum(weight1, dim=1) - torch.sum(weight2, dim=1))

        emd_score = torch.sum((Q_1).squeeze() * flow, 1)
        emd_score += energy_diff

        return emd_score, flow.view(-1, nelement_weight1, nelement_weight2)

    def forward(
        self, jets1: Tensor, jets2: Tensor, return_flows: bool = False
    ) -> Union[Tensor, Tuple[Tensor, Tensor]]:
        """
        Calculate EMD between ``jets1`` and ``jets2``.

        Args:
            jets1 (Tensor): tensor of shape ``[num_jets, num_particles, num_features]``, with features in order ``[eta, phi, pt]``.
            jets2 (Tensor): tensor of same format as ``jets1``.
            return_flows (bool): return energy flows between particles in each jet. Defaults to False.

        Returns:
            Union[Tensor, Tuple[Tensor, Tensor]]:
            - **Tensor**: EMD scores tensor of shape [num_jets].
            - **Tensor** *Optional*, if ``return_flows`` is True: tensor of flows between particles of shape
              ``[num_jets, num_particles, num_particles]``.

        """
        assert (len(jets1.shape) == 3) and (len(jets2.shape) == 3), "Jets shape incorrect"
        assert jets1.shape[0] == jets2.shape[0], "jets1 and jets2 have different numbers of jets"
        assert (jets1.shape[1] == self.num_particles) and (
            jets2.shape[1] == self.num_particles
        ), "jets don't have num_particles particles"

        if self.method == "cvxpy":
            diffs = -(jets1[:, :, :2].unsqueeze(2) - jets2[:, :, :2].unsqueeze(1)) + 1e-12
            dists = torch.norm(diffs, dim=3).view(-1, self.num_particles * self.num_particles)

            weights = torch.cat((jets1[:, :, 2], jets2[:, :, 2]), dim=1)

            E1 = torch.sum(jets1[:, :, 2], dim=1)
            E2 = torch.sum(jets2[:, :, 2], dim=1)

            Emin = torch.minimum(E1, E2).unsqueeze(1)
            EabsDiff = torch.abs(E2 - E1).unsqueeze(1)

            (flows,) = self.cvxpylayer(dists, weights, Emin)

            emds = torch.sum(dists * flows, dim=1) + EabsDiff
        elif self.method == "qpth":
            diffs = -(jets1[:, :, :2].unsqueeze(2) - jets2[:, :, :2].unsqueeze(1)) + 1e-12
            dists = torch.norm(diffs, dim=3)

            emds, flows = self._emd_inference_qpth(dists, jets1[:, :, 2], jets2[:, :, 2])

        return (emds, flows) if return_flows else emds



# __init__

# IMPORTANT: evaluation has to be imported first since energyflow must be imported before torch
# See https://github.com/pkomiske/EnergyFlow/issues/24
import jetnet.evaluation
import jetnet.datasets
import jetnet.losses
import jetnet.utils


__version__ = "0.1.3"



from typing import List, Union, Optional

import torch
from torch import Tensor
import numpy as np

import logging
from os.path import exists


# TODO: allow for loading all three jet types together


class JetNet(torch.utils.data.Dataset):
    """
    PyTorch ``torch.utils.data.Dataset`` class for the JetNet dataset.

    Features, in order: ``[eta, phi, pt, mask]``.

    Will produce an iteratable of either the dataset alone, of shape
    ``[num_jets, num_particles, num_features]``, or a tuple of the dataset and jet-level features
    of each jet. Currently only the number of (non-zero-padded) particles per jet is available as
    a jet feature.

    If pt or hdf5 files are not found in the ``data_dir`` directory then:
    If ``num_particles <= 30``, JetNet is downloaded from https://zenodo.org/record/6302454;
    Else, JetNet150 is downloaded from https://zenodo.org/record/6302240

    Args:
        jet_type (str): 'g' (gluon), 't' (top quarks), or 'q' (light quarks).
        data_dir (str): directory which contains (or in which to download) dataset.
          Defaults to "./" i.e. the working directory.
        download (bool): download the dataset, even if the hdf5 file exists already.
          Defaults to False.
        num_particles (int): number of particles to use, has to be less than or equal to 150.
          Defaults to 30.
        normalize (bool): normalize features for training or not, using parameters defined below.
          Defaults to True.
        feature_norms (Union[float, List[float]]): max value to scale each feature to.
          Can either be a single float for all features, or a list of length ``num_features``.
          Defaults to 1.0.
        feature_shifts (Union[float, List[float]]): after scaling, value to shift feature by.
          Can either be a single float for all features, or a list of length ``num_features``.
          Defaults to 0.0.
        use_mask (bool): Defaults to True.
        train (bool): whether for training or testing. Defaults to True.
        train_fraction (float): fraction of data to use as training - rest is for testing.
          Defaults to 0.7.
        num_pad_particles (int): how many out of ``num_particles`` should be zero-padded.
          Defaults to 0.
        use_num_particles_jet_feature (bool): Store the # of particles in each jet as a
          jet-level feature. *Only works if using mask* i.e. if ``use_mask=True``. Defaults to True.
        noise_padding (bool): instead of 0s, pad extra particles with Gaussian noise.
          Only works if using mask. Defaults to False.
    """

    _num_non_mask_features = 3

    # normalization used for ParticleNet training
    _fpnd_feature_maxes = [1.6211985349655151, 0.520724892616272, 0.8934717178344727, 1.0]
    _fpnd_feature_norms = 1.0
    _fpnd_feature_shifts = [0.0, 0.0, -0.5, 0.0]

    def __init__(
        self,
        jet_type: str,
        data_dir: str = "./",
        download: bool = False,
        num_particles: int = 30,
        normalize: bool = True,
        feature_norms: List[float] = [1.0, 1.0, 1.0, 1.0],
        feature_shifts: List[float] = [0.0, 0.0, -0.5, -0.5],
        use_mask: bool = True,
        train: bool = True,
        train_fraction: float = 0.7,
        num_pad_particles: int = 0,
        use_num_particles_jet_feature: bool = True,
        noise_padding: bool = False,
    ):
        assert jet_type in ["g", "t", "q", "w", "z"], "Invalid jet type"

        self.feature_norms = feature_norms
        self.feature_shifts = feature_shifts
        self.use_mask = use_mask
        # in the future there'll be more jet features such as jet pT and eta
        self.use_jet_features = use_num_particles_jet_feature and self.use_mask
        self.noise_padding = noise_padding and self.use_masks
        self.normalize = normalize

        # Use JetNet150 if ``num_particles`` > 30
        use_150 = num_particles > 30
        pt_file = f"{data_dir}/{jet_type}{'150' if use_150 else ''}.pt"

        if not exists(pt_file) or download:
            self.download_and_convert_to_pt(data_dir, jet_type, use_150)

        logging.info("Loading dataset")
        dataset = self.load_dataset(pt_file, num_particles, num_pad_particles, use_mask)
        self.num_particles = num_particles if num_particles > 0 else dataset.shape[1]

        if self.use_jet_features:
            jet_features = self.get_jet_features(dataset, use_num_particles_jet_feature)

        logging.info(f"Loaded dataset with shape {dataset.shape}")
        if normalize:
            logging.info("Normalizing features")
            self.feature_maxes = self.normalize_features(dataset, feature_norms, feature_shifts)

        if self.noise_padding:
            dataset = self.add_noise_padding(dataset)

        tcut = int(len(dataset) * train_fraction)

        self.data = dataset[:tcut] if train else dataset[tcut:]
        if self.use_jet_features:
            self.jet_features = jet_features[:tcut] if train else jet_features[tcut:]

        logging.info("Dataset processed")

    def download_and_convert_to_pt(self, data_dir: str, jet_type: str, use_150: bool = False):
        """
        Download jet dataset and convert and save to pytorch tensor.

        Args:
            data_dir (str): directory in which to save file.
            jet_type (str): jet type to download, out of ``['g', 't', 'q']``.
            use_150 (bool): download JetNet150 or JetNet. Defaults to False.

        """
        import os

        os.system(f"mkdir -p {data_dir}")
        hdf5_file = f"{data_dir}/{jet_type}{'150' if use_150 else ''}.hdf5"

        if not exists(hdf5_file):
            logging.info(f"Downloading {jet_type} jets hdf5")
            self.download(jet_type, hdf5_file, use_150)

        logging.info(f"Converting {jet_type} jets hdf5 to pt")
        self.hdf5_to_pt(data_dir, jet_type, hdf5_file, use_150)

    def download(self, jet_type: str, hdf5_file: str, use_150: bool = False):
        """
        Downloads the ``jet_type`` jet hdf5 from Zenodo and saves it as ``hdf5_file``.

        Args:
            jet_type (str): jet type to download, out of ``['g', 't', 'q']``.
            hdf5_file (str): path to save hdf5 file.
            use_150 (bool): download JetNet150 or JetNet. Defaults to False.

        """
        import requests
        import sys

        record_id = 6302240 if use_150 else 6975118
        records_url = f"https://zenodo.org/api/records/{record_id}"
        r = requests.get(records_url).json()
        key = f"{jet_type}{'150' if use_150 else ''}.hdf5"

        # finding the url for the particular jet type dataset
        file_url = next(item for item in r["files"] if item["key"] == key)["links"]["self"]
        logging.info(f"file url: {file_url}")

        # modified from https://sumit-ghosh.com/articles/python-download-progress-bar/
        with open(hdf5_file, "wb") as f:
            response = requests.get(file_url, stream=True)
            total = response.headers.get("content-length")

            if total is None:
                f.write(response.content)
            else:
                downloaded = 0
                total = int(total)

                print("Downloading dataset")
                for data in response.iter_content(chunk_size=max(int(total / 1000), 1024 * 1024)):
                    downloaded += len(data)
                    f.write(data)
                    done = int(50 * downloaded / total)
                    sys.stdout.write(
                        "\r[{}{}] {:.0f}%".format(
                            "█" * done, "." * (50 - done), float(downloaded / total) * 100
                        )
                    )
                    sys.stdout.flush()

        sys.stdout.write("\n")

    def hdf5_to_pt(self, data_dir: str, jet_type: str, hdf5_file: str, use_150: bool = False):
        """
        Converts and saves downloaded hdf5 file to pytorch tensor.

        Args:
            data_dir (str): directory in which to save file.
            jet_type (str): jet type to download, out of ``['g', 't', 'q']``.
            hdf5_file (str): path to hdf5 file.
            use_150 (bool): download JetNet150 or JetNet. Defaults to False.

        """
        import h5py

        pt_file = f"{data_dir}/{jet_type}{'150' if use_150 else ''}.pt"

        with h5py.File(hdf5_file, "r") as f:
            torch.save(Tensor(np.array(f["particle_features"])), pt_file)

    def load_dataset(
        self, pt_file: str, num_particles: int, num_pad_particles: int = 0, use_mask: bool = True
    ) -> Tensor:
        """
        Load the dataset, optionally padding the particles.

        Args:
            pt_file (str): path to dataset .pt file.
            num_particles (int): number of particles per jet to load
              (has to be less than the number per jet in the dataset).
            num_pad_particles (int): out of ``num_particles`` how many are to be zero-padded.
              Defaults to 0.
            use_mask (bool): keep or remove the mask feature. Defaults to True.

        Returns:
            Tensor: dataset tensor of shape ``[num_jets, num_particles, num_features]``.

        """
        dataset = torch.load(pt_file).float()

        # only retain up to ``num_particles``,
        # subtracting ``num_pad_particles`` since they will be padded below
        if 0 < num_particles - num_pad_particles < dataset.shape[1]:
            dataset = dataset[:, : num_particles - num_pad_particles, :]

        # pad with ``num_pad_particles`` particles
        if num_pad_particles > 0:
            dataset = torch.nn.functional.pad(dataset, (0, 0, 0, num_pad_particles), "constant", 0)

        if not use_mask:
            # remove mask feature from dataset if not needed
            dataset = dataset[:, :, : self._num_non_mask_features]

        return dataset

    def get_jet_features(self, dataset: Tensor, use_num_particles_jet_feature: bool) -> Tensor:
        """
        Returns jet-level features. `Will be expanded to include jet pT and eta.`

        Args:
            dataset (Tensor):  dataset tensor of shape [N, num_particles, num_features],
              where the last feature is the mask.
            use_num_particles_jet_feature (bool): `Currently does nothing,
              in the future such bools will specify which jet features to use`.

        Returns:
            Tensor: jet features tensor of shape [N, num_jet_features].

        """
        jet_num_particles = (torch.sum(dataset[:, :, -1], dim=1) / self.num_particles).unsqueeze(1)
        logging.debug(f"# particles: {self.num_particles}")
        return jet_num_particles

    @classmethod
    def normalize_features(
        self,
        dataset: Tensor,
        feature_norms: Union[float, List[float]] = 1.0,
        feature_shifts: Union[float, List[float]] = 0.0,
        fpnd: bool = False,
    ) -> Optional[List]:
        """
        Normalizes dataset features (in place),
        by scaling to ``feature_norms`` maximum and shifting by ``feature_shifts``.

        If the value in the List for a feature is None, it won't be scaled or shifted.

        If ``fpnd`` is True, will normalize instead to the same scale as was used for the
        ParticleNet training in https://arxiv.org/abs/2106.11535.

        Args:
            dataset (Tensor): dataset tensor of shape [N, num_particles, num_features].
            feature_norms (Union[float, List[float]]): max value to scale each feature to.
              Can either be a single float for all features, or a list of length ``num_features``.
              Defaults to 1.0.
            feature_shifts (Union[float, List[float]]): after scaling, value to shift feature by.
              Can either be a single float for all features, or a list of length ``num_features``.
              Defaults to 0.0.
            fpnd (bool): Normalize features for ParticleNet inference for the
              Frechet ParticleNet Distance metric. Will override `feature_norms`` and
              ``feature_shifts`` inputs. Defaults to False.

        Returns:
            Optional[List]: if ``fpnd`` is False, returns list of length ``num_features``
            of max absolute values for each feature. Used for unnormalizing features.

        """
        num_features = dataset.shape[2]

        if not fpnd:
            feature_maxes = [
                float(torch.max(torch.abs(dataset[:, :, i]))) for i in range(num_features)
            ]
        else:
            feature_maxes = JetNet._fpnd_feature_maxes
            feature_norms = JetNet._fpnd_feature_norms
            feature_shifts = JetNet._fpnd_feature_shifts

        if isinstance(feature_norms, float):
            feature_norms = np.full(num_features, feature_norms)

        if isinstance(feature_shifts, float):
            feature_shifts = np.full(num_features, feature_shifts)

        logging.debug(f"feature maxes: {feature_maxes}")

        for i in range(num_features):
            if feature_norms[i] is not None:
                dataset[:, :, i] /= feature_maxes[i]
                dataset[:, :, i] *= feature_norms[i]

            if feature_shifts[i] is not None and feature_shifts[i] != 0:
                dataset[:, :, i] += feature_shifts[i]

        if not fpnd:
            return feature_maxes

    def unnormalize_features(
        self,
        dataset: Union[Tensor, np.ndarray],
        ret_mask_separate: bool = True,
        is_real_data: bool = False,
        zero_mask_particles: bool = True,
        zero_neg_pt: bool = True,
    ):
        """
        Inverts the ``normalize_features()`` function on the input ``dataset`` array or tensor,
        plus optionally zero's the masked particles and negative pTs.
        Only applicable if dataset was normalized first
        i.e. ``normalize`` arg into JetNet instance is True.

        Args:
            dataset (Union[Tensor, np.ndarray]): Dataset to unnormalize.
            ret_mask_separate (bool): Return the jet and mask separately. Defaults to True.
            is_real_data (bool): Real or generated data. Defaults to False.
            zero_mask_particles (bool): Set features of zero-masked particles to 0.
              Not needed for real data. Defaults to True.
            zero_neg_pt (bool): Set pT to 0 for particles with negative pt.
              Not needed for real data. Defaults to True.

        Returns:
            Unnormalized dataset of same type as input. Either a tensor/array of shape
            ``[num_jets, num_particles, num_features (including mask)]`` if ``ret_mask_separate``
            is False, else a tuple with a tensor/array of shape
            ``[num_jets, num_particles, num_features (excluding mask)]`` and another binary mask
            tensor/array of shape ``[num_jets, num_particles, 1]``.
        """
        if not self.normalize:
            raise RuntimeError("Can't unnormalize features if dataset has not been normalized.")

        num_features = dataset.shape[2]

        for i in range(num_features):
            if self.feature_shifts[i] is not None and self.feature_shifts[i] != 0:
                dataset[:, :, i] -= self.feature_shifts[i]

            if self.feature_norms[i] is not None:
                dataset[:, :, i] /= self.feature_norms[i]
                dataset[:, :, i] *= self.feature_maxes[i]

        mask = dataset[:, :, -1] >= 0.5 if self.use_mask else None

        if not is_real_data and zero_mask_particles and self.use_mask:
            dataset[~mask] = 0

        if not is_real_data and zero_neg_pt:
            dataset[:, :, 2][dataset[:, :, 2] < 0] = 0

        return dataset[:, :, : self._num_non_mask_features], mask if ret_mask_separate else dataset

    def add_noise_padding(self, dataset: Tensor):
        """Add Gaussian noise to zero-masked particles"""
        # up to 5 sigmas will be within ±1
        noise_padding = torch.randn((len(dataset), self.num_particles, dataset.shape[2] - 1)) / 5
        noise_padding[noise_padding > 1] = 1
        noise_padding[noise_padding < -1] = -1
        noise_padding[:, :, 2] /= 2.0  # pt is scaled between ±0.5

        mask = (dataset[:, :, -1] + 0.5).bool()
        noise_padding[mask] = 0  # only adding noise to zero-masked particles
        dataset += torch.cat(
            (noise_padding, torch.zeros((len(dataset), self.num_particles, 1))), dim=2
        )

        return dataset

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        return (self.data[idx], self.jet_features[idx]) if self.use_jet_features else self.data[idx]



from typing import Dict, Union

import numpy as np

# for calculating jet features quickly,
# TODO: replace with vector library when summing over axis feature is implemented
import awkward as ak
from coffea.nanoevents.methods import vector

from energyflow import EFPSet

ak.behavior.update(vector.behavior)


def jet_features(jets: np.ndarray) -> Dict[str, Union[float, np.ndarray]]:
    """
    Calculates jet features by summing over particle Lorentz 4-vectors.

    Args:
        jets (np.ndarray): array of either a single or multiple jets, of shape either
          ``[num_particles, num_features]`` or ``[num_jets, num_particles, num_features]``,
          with features in order ``[eta, phi, pt, (optional) mass]``. If no particle masses given,
          they are assumed to be 0.

    Returns:
        Dict[str, Union[float, np.ndarray]]: dict of float (if inputted single jet) or
          1D arrays of length ``num_jets`` (if inputted multiple jets)
          with 'mass', 'pt', and 'eta' keys.

    """

    assert len(jets.shape) == 2 or len(jets.shape) == 3, "jets dimensions are incorrect"
    assert jets.shape[-1] >= 3, "missing particle features"

    if len(jets.shape) == 2:
        vecs = ak.zip(
            {
                "pt": jets[:, 2:3],
                "eta": jets[:, 0:1],
                "phi": jets[:, 1:2],
                # 0s for mass if no mass given
                "mass": ak.full_like(jets[:, 2:3], 0) if jets.shape[1] == 3 else jets[:, 3:4],
            },
            with_name="PtEtaPhiMLorentzVector",
        )

        sum_vecs = vecs.sum(axis=0)
    else:
        vecs = ak.zip(
            {
                "pt": jets[:, :, 2:3],
                "eta": jets[:, :, 0:1],
                "phi": jets[:, :, 1:2],
                # 0s for mass if no mass given
                "mass": ak.full_like(jets[:, :, 2:3], 0) if jets.shape[2] == 3 else jets[:, :, 3:4],
            },
            with_name="PtEtaPhiMLorentzVector",
        )

        sum_vecs = vecs.sum(axis=1)

    jf = {
        "mass": np.nan_to_num(np.array(sum_vecs.mass)).squeeze(),
        "pt": np.nan_to_num(np.array(sum_vecs.pt)).squeeze(),
        "eta": np.nan_to_num(np.array(sum_vecs.eta)).squeeze(),
    }

    return jf


def efps(
    jets: np.ndarray,
    use_particle_masses: bool = False,
    efpset_args: list = [("n==", 4), ("d==", 4), ("p==", 1)],
    efp_jobs: int = None,
) -> np.ndarray:
    """
    Utility for calculating EFPs for jets in JetNet format using the energyflow library.

    Args:
        jets (np.ndarray): array of either a single or multiple jets, of shape either
          ``[num_particles, num_features]`` or ``[num_jets, num_particles, num_features]``,
          with features in order ``[eta, phi, pt, (optional) mass]``. If no particle masses given,
          they are assumed to be 0.
        efpset_args (List): Args for the energyflow.efpset function to specify which EFPs to use,
          as defined here https://energyflow.network/docs/efp/#efpset.
          Defaults to the n=4, d=5, prime EFPs.
        efp_jobs (int): number of jobs to use for energyflow's EFP batch computation.
          None means as many processes as there are CPUs.

    Returns:
        np.ndarray: 1D (if inputted single jet) or 2D array of shape ``[num_jets, num_efps]`` of
          EFPs per jet

    """

    assert len(jets.shape) == 2 or len(jets.shape) == 3, "jets dimensions are incorrect"
    assert jets.shape[-1] - int(use_particle_masses) >= 3, "particle feature format is incorrect"

    efpset = EFPSet(*efpset_args, measure="hadr", beta=1, normed=None, coords="ptyphim")

    if len(jets.shape) == 2:
        # convert to energyflow format
        jets = jets[:, [2, 0, 1]] if not use_particle_masses else jets[:, [2, 0, 1, 3]]
        efps = efpset.compute(jets)
    else:
        # convert to energyflow format
        jets = jets[:, :, [2, 0, 1]] if not use_particle_masses else jets[:, :, [2, 0, 1, 3]]
        efps = efpset.batch_compute(jets, efp_jobs)

    return efps


def to_image(
    jet: np.ndarray, im_size: int, mask: np.ndarray = None, maxR: float = 1.0
) -> np.ndarray:
    """
    Convert a single jet into a 2D ``im_size`` x ``im_size`` array.

    Args:
        jet (np.ndarray): array of a single jet of shape ``[num_particles, num_features]``
          with features in order ``[eta, phi, pt]``.
        im_size (int): number of pixels per row and column.
        mask (np.ndarray): optional binary array of masks of shape ``[num_particles]``.
        maxR (float): max radius of the jet. Defaults to 1.0.

    Returns:
        np.ndarray: 2D array of shape ``[im_size, im_size]``.

    """
    assert len(jet.shape) == 2, "jets dimensions are incorrect"
    assert jet.shape[-1] >= 3, "particle feature format is incorrect"

    bins = np.linspace(-maxR, maxR, im_size + 1)
    binned_eta = np.digitize(jet[:, 0], bins) - 1
    binned_phi = np.digitize(jet[:, 1], bins) - 1
    pt = jet[:, 2]

    if mask is not None:
        assert len(mask.shape) == 1 and mask.shape[0] == jet.shape[0], "mask format incorrect"
        mask = mask.astype(int)
        pt *= mask

    jet_image = np.zeros((im_size, im_size))

    for eta, phi, pt in zip(binned_eta, binned_phi, pt):
        if eta >= 0 and eta < im_size and phi >= 0 and phi < im_size:
            jet_image[phi, eta] += pt

    return jet_image



import torch
from torch import nn

from torch_geometric.nn import EdgeConv, global_mean_pool
from torch_cluster import knn_graph

import numpy as np
import torch.nn.functional as F


class _ParticleNetEdgeNet(nn.Module):
    def __init__(self, in_size, layer_size):
        super(_ParticleNetEdgeNet, self).__init__()

        layers = []

        layers.append(nn.Linear(in_size * 2, layer_size))
        layers.append(nn.BatchNorm1d(layer_size))
        layers.append(nn.ReLU())

        for i in range(2):
            layers.append(nn.Linear(layer_size, layer_size))
            layers.append(nn.BatchNorm1d(layer_size))
            layers.append(nn.ReLU())

        self.model = nn.Sequential(*layers)

    def forward(self, x):
        return self.model(x)

    def __repr__(self):
        return "{}(nn={})".format(self.__class__.__name__, self.model)


class _ParticleNet(nn.Module):
    def __init__(self, num_hits, node_feat_size, num_classes=5):
        super(_ParticleNet, self).__init__()
        self.num_hits = num_hits
        self.node_feat_size = node_feat_size
        self.num_classes = num_classes

        self.k = 16
        self.num_edge_convs = 3
        self.kernel_sizes = [64, 128, 256]
        self.fc_size = 256
        self.dropout = 0.1

        self.edge_nets = nn.ModuleList()
        self.edge_convs = nn.ModuleList()

        self.kernel_sizes.insert(0, self.node_feat_size)
        self.output_sizes = np.cumsum(self.kernel_sizes)

        self.edge_nets.append(_ParticleNetEdgeNet(self.node_feat_size, self.kernel_sizes[1]))
        self.edge_convs.append(EdgeConv(self.edge_nets[-1], aggr="mean"))

        for i in range(1, self.num_edge_convs):
            # adding kernel sizes because of skip connections
            self.edge_nets.append(
                _ParticleNetEdgeNet(self.output_sizes[i], self.kernel_sizes[i + 1])
            )
            self.edge_convs.append(EdgeConv(self.edge_nets[-1], aggr="mean"))

        self.fc1 = nn.Sequential(nn.Linear(self.output_sizes[-1], self.fc_size))

        self.dropout_layer = nn.Dropout(p=self.dropout)

        self.fc2 = nn.Linear(self.fc_size, self.num_classes)

    def forward(self, x, ret_activations=False, relu_activations=False):
        batch_size = x.size(0)
        x = x.reshape(batch_size * self.num_hits, self.node_feat_size)
        zeros = torch.zeros(batch_size * self.num_hits, dtype=int).to(x.device)
        zeros[torch.arange(batch_size) * self.num_hits] = 1
        batch = torch.cumsum(zeros, 0) - 1

        for i in range(self.num_edge_convs):
            # using only angular coords for knn in first edgeconv block
            edge_index = (
                knn_graph(x[:, :2], self.k, batch) if i == 0 else knn_graph(x, self.k, batch)
            )
            x = torch.cat(
                (self.edge_convs[i](x, edge_index), x), dim=1
            )  # concatenating with original features i.e. skip connection

        x = global_mean_pool(x, batch)
        x = self.fc1(x)

        if ret_activations:
            if relu_activations:
                return F.relu(x)
            else:
                return x  # for Frechet ParticleNet Distance
        else:
            x = self.dropout_layer(F.relu(x))

        return self.fc2(x)  # no softmax because pytorch cross entropy loss includes softmax

    # TODO: !!
    # def __repr__(self):
    #     return ""



from setuptools import setup, find_packages
import re

install_requires = [
    "numpy >= 1.21.0",
    "torch >= 1.8.0",
    "energyflow >= 1.3.0",
    "scipy >= 1.6.2",
    "awkward >= 1.4.0",
    "coffea >= 0.7.0",
    "h5py >= 3.0.0",
    "requests",
    "tqdm",
]

extras_require = {"emdloss": ["qpth", "cvxpy"]}


def readme():
    with open("README.md") as f:
        return f.read()


with open("jetnet/__init__.py", "r") as f:
    __version__ = re.search(r'__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read()).group(1)


setup(
    name="jetnet",
    version=__version__,
    description="Jets + ML integration",
    long_description=readme(),
    long_description_content_type="text/markdown",
    url="http://github.com/jet-net/JetNet",
    author="Raghav Kansal",
    author_email="rkansal@cern.ch",
    license="MIT",
    packages=find_packages(),
    install_requires=install_requires,
    extras_require=extras_require,
    zip_safe=False,
    include_package_data=True,
)
